import { WebhookHandler } from "@webuildbots/webuildbots-sdk/lib/util/webhook-client";
import {
  BasicBB,
  BlockPointer,
  ResponseBuilder,
  WebhookRequest,
} from "@webuildbots/webuildbots-sdk";
import { backToMenu } from "@webuildbots/webuildbots-sdk/lib/builder/choice/common-cb";
import GoogleSheetsService from "../services/google-sheets.service";
import { FunctionFailureParams } from "@webuildbots/webuildbots-sdk/lib/interfaces/webhook";
import { getPathValuesFromForm } from "../util/form-value-to-spreadsheet";
import { GoogleSpreadsheetRow } from "google-spreadsheet";

// error blocks
enum SpreedsheetBlock {
  CREDENTIALS_NOT_SET = "no-google-credentials-set",
  SPREADSHEET_ID_NOT_SET = "no-spreadsheet-id-set",
  KEY_COLUMN_NAME_NOT_SET = "no-key-column-name-set",
  RETURN_STRING_NOT_SET = "no-return-string-set",
  VALUE_PROPERTY_NAME_NOT_SET = "no-value-property-name-set",
}

const checkGoogleCredentials = (credentials: {
  clientEmail: string;
  privateKey: string;
}): FunctionFailureParams | null => {
  if (!credentials?.clientEmail || !credentials?.privateKey) {
    const errorBlock: BlockPointer = {
      id: SpreedsheetBlock.CREDENTIALS_NOT_SET,
    };
    return {
      errorBP: errorBlock,
      resetForm: true,
      tryAgain: false,
    };
  }
};

const checkSpreadsheetArgs = (
  param: string,
  blockId: string
): FunctionFailureParams | null => {
  if (!param) {
    const errorBlock: BlockPointer = { id: blockId };
    return {
      errorBP: errorBlock,
      resetForm: true,
      tryAgain: false,
    };
  }
};

const parseReturnString = (
  returnString: string,
  requstedRow: GoogleSpreadsheetRow
): string => {
  // Replace any templates (e.g. %%username%%) with appropriate row data
  const matches = returnString.match(/(%%[-a-zA-Z\d. ]+%%)/g) || [];
  for (const match of matches.map((s) => s.substring(2, s.length - 2))) {
    const templateString = "%%" + match + "%%";
    returnString = returnString.replace(templateString, requstedRow[match]);
  }
  return returnString;
};

export const getData: WebhookHandler = async (
  webhookReq: WebhookRequest,
  responseBuilder: ResponseBuilder
): Promise<void> => {
  // checking for required arguments
  const credentialsFP = checkGoogleCredentials(webhookReq.args?.credentials);
  const spreadsheetIDFP = checkSpreadsheetArgs(
    webhookReq.args?.spreadsheetID,
    SpreedsheetBlock.SPREADSHEET_ID_NOT_SET
  );
  const keyColumnNameFP = checkSpreadsheetArgs(
    webhookReq.args?.keyColumnName,
    SpreedsheetBlock.KEY_COLUMN_NAME_NOT_SET
  );
  const returnStringFP = checkSpreadsheetArgs(
    webhookReq.args?.returnString,
    SpreedsheetBlock.RETURN_STRING_NOT_SET
  );
  const valuePropertyNameFP = checkSpreadsheetArgs(
    webhookReq.args?.valuePropertyName,
    SpreedsheetBlock.VALUE_PROPERTY_NAME_NOT_SET
  );
  if (credentialsFP) {
    responseBuilder.functionFailure(credentialsFP);
  } else if (spreadsheetIDFP) {
    responseBuilder.functionFailure(spreadsheetIDFP);
  } else if (keyColumnNameFP) {
    responseBuilder.functionFailure(keyColumnNameFP);
  } else if (returnStringFP) {
    responseBuilder.functionFailure(returnStringFP);
  } else if (valuePropertyNameFP) {
    responseBuilder.functionFailure(valuePropertyNameFP);
  } else {
    const {
      valuePropertyName,
      credentials,
      keyColumnName,
      spreadsheetID,
      returnString,
      sheetIndex,
    } = webhookReq.args;
    const { privateKey, clientEmail } = credentials;
    // get form value
    const { formValue, user } = webhookReq;
    const [pathValue] = getPathValuesFromForm(formValue, user).filter(
      (item) => item.path.toLowerCase() === valuePropertyName.toLowerCase()
    );
    // find requested row
    const googleSheetService = new GoogleSheetsService(
      spreadsheetID,
      clientEmail,
      privateKey
    );
    const rows = await googleSheetService.getRows(sheetIndex);
    const requestedRow = rows.filter(
      (item) => `${item[keyColumnName]}` === `${pathValue.value}`
    );
    if (requestedRow.length) {
      // build block from returnString
      const text = parseReturnString(returnString, requestedRow[0]);
      const foundRowResponse = new BasicBB()
        .addText(webhookReq.user.locality.language, text)
        .pushChoice(backToMenu());
      responseBuilder.pushBlock(foundRowResponse).unsetFunctionForm();
    } else {
      // send not found data block
      responseBuilder
        .pushBlockPointer({ id: webhookReq.args?.notFoundDataBlock })
        .unsetFunctionForm();
    }
  }
};

export const saveData: WebhookHandler = async (
  webhookReq: WebhookRequest,
  responseBuilder: ResponseBuilder
): Promise<void> => {
  const credentialsFP = checkGoogleCredentials(webhookReq.args?.credentials);
  const spreadsheetIDFP = checkSpreadsheetArgs(
    webhookReq.args?.spreadsheetID,
    SpreedsheetBlock.SPREADSHEET_ID_NOT_SET
  );

  if (credentialsFP) {
    responseBuilder.functionFailure(credentialsFP);
  } else if (spreadsheetIDFP) {
    responseBuilder.functionFailure(spreadsheetIDFP);
  } else {
    const { credentials, spreadsheetID, saveUserId } = webhookReq.args;
    const { privateKey, clientEmail } = credentials;
    const googleSheetService = new GoogleSheetsService(
      spreadsheetID,
      clientEmail,
      privateKey
    );

    const { formValue, user } = webhookReq;
    const pathValues = getPathValuesFromForm(formValue, user, saveUserId);
    await googleSheetService.addRow(pathValues);

    responseBuilder.pushBlockPointer({
      id: webhookReq.args?.confirmationBlock,
    });
  }
};
