import {
  GoogleSpreadsheet,
  GoogleSpreadsheetRow,
  GoogleSpreadsheetWorksheet,
} from "google-spreadsheet";
import isString from "lodash/isString";

/**
 * Used grant programmatic access to google sheets docs.
 *
 * N.B. if you are getting 403 errors then your may need to share the sheet with
 * the service user you are authenticating with. Check the credentials at config.
 */
export default class GoogleSheetsService {
  private sheets: GoogleSpreadsheet;

  /**
   * @param spreadsheetID - a spreadsheet key is the long id in the sheets URL
   * @param clientEmail - client_email is copied from service account credentials generated by google
   * @param privateKey - private_key is copied from service account credentials generated by google
   */
  constructor(
    private spreadsheetID: string,
    private clientEmail: string,
    private privateKey: string
  ) {}

  public async addRow(
    rows: { path: string; value: string }[],
    sheetNumber = 0
  ): Promise<void> {
    const sheet = await this.getSheet(sheetNumber);
    if (!sheet.headerValues) {
      const headers = rows.map((row) => row.path.toLowerCase());
      await sheet.setHeaderRow(headers);
    }
    const insertRow = {};
    rows.forEach((row) => {
      const rowValue = isString(row.value) ? row.value.trim(): row.value;
      insertRow[row.path.toLowerCase().trim()] = rowValue;
    });
    await sheet.addRow(insertRow);
  }

  public async getSheets(): Promise<GoogleSpreadsheet> {
    if (!this.sheets) {
      // spreadsheet key is the long id in the sheets URL
      this.sheets = new GoogleSpreadsheet(this.spreadsheetID);
      // use service account creds load directly from json file
      await this.sheets.useServiceAccountAuth({
        private_key: this.privateKey,
        client_email: this.clientEmail,
      });
    }
    return this.sheets;
  }

  /**
   * @param sheetNumber - sheet number starting from 1
   */
  public async getSheet(
    sheetNumber: number
  ): Promise<GoogleSpreadsheetWorksheet> {
    const sheets = await this.getSheets();
    // loads document properties and worksheets
    await sheets.loadInfo();
    return sheets.sheetsByIndex[sheetNumber];
  }

  public async getRows(sheetNumber = 0): Promise<GoogleSpreadsheetRow[]> {
    const sheet = await this.getSheet(sheetNumber);
    return sheet.getRows();
  }
}
